"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from synalinks.src.ops.knowledge_bases import similarity_search as similarity_search
from synalinks.src.ops.knowledge_bases import update_knowledge as update_knowledge
