"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from synalinks.src.programs.program import Program as Program
from synalinks.src.programs.program import program_from_json as program_from_json
from synalinks.src.programs.sequential import Sequential as Sequential
